@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.157.2.14 2005/01/20 17:01:08 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/main.c,v 1.198.2.20 2006/02/17 07:38:40 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2005 The OpenLDAP Foundation.
+ * Copyright 1998-2006 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,6 +27,7 @@
 
 #include <stdio.h>
 
+#include <ac/ctype.h>
 #include <ac/socket.h>
 #include <ac/string.h>
 #include <ac/time.h>
@@ -34,8 +35,6 @@
 #include <ac/wait.h>
 #include <ac/errno.h>
 
-#include "ldap_pvt.h"
-
 #include "slap.h"
 #include "lutil.h"
 #include "ldif.h"
@@ -65,7 +64,8 @@ static struct sockaddr_in	bind_addr;
 #endif
 
 typedef int (MainFunc) LDAP_P(( int argc, char *argv[] ));
-extern MainFunc slapadd, slapcat, slapdn, slapindex, slappasswd, slaptest;
+extern MainFunc slapadd, slapcat, slapdn, slapindex, slappasswd,
+	slaptest, slapauth, slapacl;
 
 static struct {
 	char *name;
@@ -77,6 +77,14 @@ static struct {
 	{"slapindex", slapindex},
 	{"slappasswd", slappasswd},
 	{"slaptest", slaptest},
+	{"slapauth", slapauth},
+	{"slapacl", slapacl},
+	/* NOTE: new tools must be added in chronological order,
+	 * not in alphabetical order, because for backwards
+	 * compatibility name[4] is used to identify the
+	 * tools; so name[4]=='a' must refer to "slapadd" and
+	 * not to "slapauth".  Alphabetical order can be used
+	 * for tools whose name[4] is not used yet */
 	{NULL, NULL}
 };
 
@@ -128,6 +136,120 @@ static int version = 0;
 
 void *slap_tls_ctx;
 
+static int
+slapd_opt_slp( const char *val, void *arg )
+{
+#ifdef HAVE_SLP
+	/* NULL is default */
+	if ( val == NULL || strcasecmp( val, "on" ) == 0 ) {
+		slapd_register_slp = 1;
+
+	} else if ( strcasecmp( val, "off" ) == 0 ) {
+		slapd_register_slp = 0;
+
+	/* NOTE: add support for URL specification? */
+
+	} else {
+		fprintf(stderr, "unrecognized value \"%s\" for SLP option\n", val );
+		return -1;
+	}
+
+	return 0;
+		
+#else
+	fputs( "slapd: SLP support is not available\n", stderr );
+	return 0;
+#endif
+}
+
+/*
+ * Option helper structure:
+ * 
+ * oh_nam	is left-hand part of <option>[=<value>]
+ * oh_fnc	is handler function
+ * oh_arg	is an optional arg to oh_fnc
+ * oh_usage	is the one-line usage string related to the option,
+ *		which is assumed to start with <option>[=<value>]
+ *
+ * please leave valid options in the structure, and optionally #ifdef
+ * their processing inside the helper, so that reasonable and helpful
+ * error messages can be generated if a disabled option is requested.
+ */
+struct option_helper {
+	struct berval	oh_name;
+	int		(*oh_fnc)(const char *val, void *arg);
+	void		*oh_arg;
+	const char	*oh_usage;
+} option_helpers[] = {
+	{ BER_BVC("slp"),	slapd_opt_slp,	NULL, "slp[={on|off}] enable/disable SLP" },
+	{ BER_BVNULL, 0, NULL, NULL }
+};
+
+int
+parse_debug_unknowns( char **unknowns, int *levelp )
+{
+	int i, level, rc = 0;
+
+	for ( i = 0; unknowns[ i ] != NULL; i++ ) {
+		level = 0;
+		if ( str2loglevel( unknowns[ i ], &level )) {
+			fprintf( stderr,
+				"unrecognized log level \"%s\"\n", unknowns[ i ] );
+			rc = 1;
+		} else {
+			*levelp |= level;
+		}
+	}
+	return rc;
+}
+
+int
+parse_debug_level( const char *arg, int *levelp, char ***unknowns )
+{
+	int	level;
+
+	if ( arg != NULL && arg[ 0 ] != '-' && !isdigit( arg[ 0 ] ) )
+	{
+		int	i;
+		char	**levels;
+
+		levels = ldap_str2charray( arg, "," );
+
+		for ( i = 0; levels[ i ] != NULL; i++ ) {
+			level = 0;
+
+			if ( str2loglevel( levels[ i ], &level ) ) {
+				/* remember this for later */
+				ldap_charray_add( unknowns, levels[ i ] );
+				fprintf( stderr,
+					"unrecognized log level \"%s\" (deferred)\n",
+					levels[ i ] );
+			} else {
+				*levelp |= level;
+			}
+		}
+
+		ldap_charray_free( levels );
+
+	} else {
+		if ( lutil_atoix( &level, arg, 0 ) != 0 ) {
+			fprintf( stderr,
+				"unrecognized log level "
+				"\"%s\"\n", arg );
+			return 1;
+		}
+
+		if ( level == 0 ) {
+			*levelp = 0;
+
+		} else {
+			*levelp |= level;
+		}
+	}
+
+	return 0;
+}
+
 static void
 usage( char *name )
 {
@@ -136,10 +258,12 @@ usage( char *name )
 	fprintf( stderr,
 		"\t-4\t\tIPv4 only\n"
 		"\t-6\t\tIPv6 only\n"
-		"\t-T {acdipt}\tRun in Tool mode\n"
+		"\t-T {acl|add|auth|cat|dn|index|passwd|test}\n"
+		"\t\t\tRun in Tool mode\n"
 		"\t-c cookie\tSync cookie of consumer\n"
 		"\t-d level\tDebug level" "\n"
 		"\t-f filename\tConfiguration file\n"
+		"\t-F dir\tConfiguration directory\n"
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID)
 		"\t-g group\tGroup (id or name) to run as\n"
 #endif
@@ -148,14 +272,26 @@ usage( char *name )
 		"\t-l facility\tSyslog facility (default: LOCAL4)\n"
 #endif
 		"\t-n serverName\tService name\n"
+		"\t-o <opt>[=val] generic means to specify options" );
+	if ( !BER_BVISNULL( &option_helpers[0].oh_name ) ) {
+		int	i;
+
+		fprintf( stderr, "; supported options:\n" );
+		for ( i = 0; !BER_BVISNULL( &option_helpers[i].oh_name ); i++) {
+			fprintf( stderr, "\t\t%s\n", option_helpers[i].oh_usage );
+		}
+	} else {
+		fprintf( stderr, "\n" );
+	}
+	fprintf( stderr,	
 #ifdef HAVE_CHROOT
 		"\t-r directory\tSandbox directory to chroot to\n"
 #endif
 		"\t-s level\tSyslog level\n"
 #if defined(HAVE_SETUID) && defined(HAVE_SETGID)
 		"\t-u user\t\tUser (id or name) to run as\n"
-		"\t-V\t\tprint version info (-VV only)\n"
 #endif
+		"\t-V\t\tprint version info (-VV only)\n"
     );
 }
 
@@ -176,24 +312,28 @@ int main( int argc, char **argv )
 	char *sandbox = NULL;
 #endif
 #ifdef LOG_LOCAL4
-    int	    syslogUser = DEFAULT_SYSLOG_USER;
+	int syslogUser = DEFAULT_SYSLOG_USER;
 #endif
 	
 	int g_argc = argc;
 	char **g_argv = argv;
 
-#ifdef HAVE_NT_SERVICE_MANAGER
-	char		*configfile = ".\\slapd.conf";
-#else
-	char		*configfile = SLAPD_DEFAULT_CONFIGFILE;
-#endif
-	char	    *serverName;
-	int	    serverMode = SLAP_SERVER_MODE;
+	char *configfile = NULL;
+	char *configdir = NULL;
+	char *serverName;
+	int serverMode = SLAP_SERVER_MODE;
 
-	struct berval cookie = BER_BVNULL;
 	struct sync_cookie *scp = NULL;
 	struct sync_cookie *scp_entry = NULL;
 
+	char **debug_unknowns = NULL;
+	char **syslog_unknowns = NULL;
+
+	char *serverNamePrefix = "";
+	size_t	l;
+
+	int slapd_pid_file_unlink = 0, slapd_args_file_unlink = 0;
+
 #ifdef CSRIMALLOC
 	FILE *leakfile;
 	if( ( leakfile = fopen( "slapd.leak", "w" )) == NULL ) {
@@ -201,7 +341,9 @@ int main( int argc, char **argv )
 	}
 #endif
 
-	sl_mem_init();
+	slap_sl_mem_init();
+
+	(void) ldap_pvt_thread_initialize();
 
 	serverName = lutil_progname( "slapd", argc, argv );
 
@@ -218,6 +360,7 @@ int main( int argc, char **argv )
 	{
 		int *i;
 		char *newConfigFile;
+		char *newConfigDir;
 		char *newUrls;
 		char *regService = NULL;
 
@@ -228,53 +371,38 @@ int main( int argc, char **argv )
 		}
 
 		i = (int*)lutil_getRegParam( regService, "DebugLevel" );
-		if ( i != NULL ) 
-		{
+		if ( i != NULL ) {
 			slap_debug = *i;
-#ifdef NEW_LOGGING
-			lutil_log_initialize( argc, argv );
-			LDAP_LOG( SLAPD, INFO, 
-				"main: new debug level from registry is: %d\n", 
-				slap_debug, 0, 0 );
-#else
-			Debug( LDAP_DEBUG_ANY, "new debug level from registry is: %d\n", slap_debug, 0, 0 );
-#endif
+			Debug( LDAP_DEBUG_ANY,
+				"new debug level from registry is: %d\n", slap_debug, 0, 0 );
 		}
 
 		newUrls = (char *) lutil_getRegParam(regService, "Urls");
-		if (newUrls)
-		{
+		if (newUrls) {
 		    if (urls)
 			ch_free(urls);
 
 		    urls = ch_strdup(newUrls);
-#ifdef NEW_LOGGING
-		    LDAP_LOG( SLAPD, INFO, 
-				"main: new urls from registry: %s\n", urls, 0, 0 );
-#else
 		    Debug(LDAP_DEBUG_ANY, "new urls from registry: %s\n",
-			  urls, 0, 0);
-#endif
-
+				urls, 0, 0);
 		}
 
 		newConfigFile = (char*)lutil_getRegParam( regService, "ConfigFile" );
-		if ( newConfigFile != NULL ) 
-		{
+		if ( newConfigFile != NULL ) {
 			configfile = newConfigFile;
-#ifdef NEW_LOGGING
-			LDAP_LOG( SLAPD, INFO, 
-				"main: new config file from registry is: %s\n", configfile, 0, 0 );
-#else
 			Debug ( LDAP_DEBUG_ANY, "new config file from registry is: %s\n", configfile, 0, 0 );
-#endif
+		}
 
+		newConfigDir = (char*)lutil_getRegParam( regService, "ConfigDir" );
+		if ( newConfigDir != NULL ) {
+			configdir = newConfigDir;
+			Debug ( LDAP_DEBUG_ANY, "new config dir from registry is: %s\n", configdir, 0, 0 );
 		}
 	}
 #endif
 
 	while ( (i = getopt( argc, argv,
-			     "c:d:f:h:s:n:tT:V"
+			     "c:d:f:F:h:n:o:s:tT:V"
 #if LDAP_PF_INET6
 				"46"
 #endif
@@ -306,21 +434,24 @@ int main( int argc, char **argv )
 		case 'c':	/* provide sync cookie, override if exist in replica */
 			scp = (struct sync_cookie *) ch_calloc( 1,
 										sizeof( struct sync_cookie ));
-			ber_str2bv( optarg, strlen( optarg ), 1, &cookie );
-			ber_bvarray_add( &scp->octet_str, &cookie );
-			slap_parse_sync_cookie( scp );
+			ber_str2bv( optarg, 0, 1, &scp->octet_str );
+			
+			/* This only parses out the rid at this point */
+			slap_parse_sync_cookie( scp, NULL );
+
+			if ( scp->rid == -1 ) {
+				Debug( LDAP_DEBUG_ANY,
+						"main: invalid cookie \"%s\"\n",
+						optarg, 0, 0 );
+				slap_sync_cookie_free( scp, 1 );
+				goto destroy;
+			}
 
 			LDAP_STAILQ_FOREACH( scp_entry, &slap_sync_cookie, sc_next ) {
 				if ( scp->rid == scp_entry->rid ) {
-#ifdef NEW_LOGGING
-					LDAP_LOG( OPERATION, CRIT,
-							"main: duplicated replica id in cookies\n",
-							0, 0, 0 );
-#else
 					Debug( LDAP_DEBUG_ANY,
 						    "main: duplicated replica id in cookies\n",
 							0, 0, 0 );
-#endif
 					slap_sync_cookie_free( scp, 1 );
 					goto destroy;
 				}
@@ -328,23 +459,68 @@ int main( int argc, char **argv )
 			LDAP_STAILQ_INSERT_TAIL( &slap_sync_cookie, scp, sc_next );
 			break;
 
-		case 'd':	/* set debug level and 'do not detach' flag */
+		case 'd': {	/* set debug level and 'do not detach' flag */
+			int	level = 0;
+
 			no_detach = 1;
+			if ( parse_debug_level( optarg, &level, &debug_unknowns ) ) {
+				goto destroy;
+			}
 #ifdef LDAP_DEBUG
-			slap_debug |= atoi( optarg );
+			slap_debug |= level;
 #else
-			if ( atoi( optarg ) != 0 )
+			if ( level != 0 )
 				fputs( "must compile with LDAP_DEBUG for debugging\n",
 				       stderr );
 #endif
-			break;
+			} break;
 
 		case 'f':	/* read config file */
 			configfile = ch_strdup( optarg );
 			break;
 
+		case 'F':	/* use config dir */
+			configdir = ch_strdup( optarg );
+			break;
+
+		case 'o': {
+			char		*val = strchr( optarg, '=' );
+			struct berval	opt;
+			int		i;
+
+			opt.bv_val = optarg;
+			
+			if ( val ) {
+				opt.bv_len = ( val - optarg );
+				val++;
+			
+			} else {
+				opt.bv_len = strlen( optarg );
+			}
+
+			for ( i = 0; !BER_BVISNULL( &option_helpers[i].oh_name ); i++ ) {
+				if ( ber_bvstrcasecmp( &option_helpers[i].oh_name, &opt ) == 0 ) {
+					assert( option_helpers[i].oh_fnc != NULL );
+					if ( (*option_helpers[i].oh_fnc)( val, option_helpers[i].oh_arg ) == -1 ) {
+						/* we assume the option parsing helper
+						 * issues appropriate and self-explanatory
+						 * error messages... */
+						goto stop;
+					}
+					break;
+				}
+			}
+
+			if ( BER_BVISNULL( &option_helpers[i].oh_name ) ) {
+				goto unhandled_option;
+			}
+			break;
+		}
+
 		case 's':	/* set syslog level */
-			ldap_syslog = atoi( optarg );
+			if ( parse_debug_level( optarg, &ldap_syslog, &syslog_unknowns ) ) {
+				goto destroy;
+			}
 			break;
 
 #ifdef LOG_LOCAL4
@@ -389,14 +565,31 @@ int main( int argc, char **argv )
 			break;
 
 		case 'T':
-			for (i=0; tools[i].name; i++) {
-				if ( optarg[0] == tools[i].name[4] ) {
-					rc = tools[i].func(argc, argv);
-					MAIN_RETURN(rc);
+			/* try full option string first */
+			for ( i = 0; tools[i].name; i++ ) {
+				if ( strcmp( optarg, &tools[i].name[4] ) == 0 ) {
+					rc = tools[i].func( argc, argv );
+					MAIN_RETURN( rc );
 				}
 			}
+
+			/* try bits of option string (backward compatibility for single char) */
+			l = strlen( optarg );
+			for ( i = 0; tools[i].name; i++ ) {
+				if ( strncmp( optarg, &tools[i].name[4], l ) == 0 ) {
+					rc = tools[i].func( argc, argv );
+					MAIN_RETURN( rc );
+				}
+			}
+			
+			/* issue error */
+			serverName = optarg;
+			serverNamePrefix = "slap";
+			fprintf( stderr, "program name \"%s%s\" unrecognized; "
+					"aborting...\n", serverNamePrefix, serverName );
 			/* FALLTHRU */
 		default:
+unhandled_option:;
 			usage( argv[0] );
 			rc = 1;
 			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 15 );
@@ -404,14 +597,9 @@ int main( int argc, char **argv )
 		}
 	}
 
-#ifdef NEW_LOGGING
-	lutil_log_initialize( argc, argv );
-#else
-	lutil_set_debug_level( "slapd", slap_debug );
 	ber_set_option(NULL, LBER_OPT_DEBUG_LEVEL, &slap_debug);
 	ldap_set_option(NULL, LDAP_OPT_DEBUG_LEVEL, &slap_debug);
 	ldif_debug = slap_debug;
-#endif
 
 	if ( version ) {
 		fprintf( stderr, "%s\n", Versionstr );
@@ -437,11 +625,7 @@ int main( int argc, char **argv )
 #endif
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, "%s", Versionstr, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "%s", Versionstr, 0, 0 );
-#endif
 
 	if( check == CHECK_NONE && slapd_daemon_init( urls ) != 0 ) {
 		rc = 1;
@@ -474,73 +658,13 @@ int main( int argc, char **argv )
 	lutil_passwd_init();
 	slap_op_init();
 
-#ifdef SLAPD_MODULES
-	if ( module_init() != 0 ) {
-		rc = 1;
-		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 17 );
-		goto destroy;
-	}
-#endif
-
-	if ( slap_schema_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: schema initialization error\n", 0, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY,
-		    "schema initialization error\n",
-		    0, 0, 0 );
-#endif
-
-		goto destroy;
-	}
-
-	if ( slap_init( serverMode, serverName ) != 0 ) {
-		rc = 1;
+	rc = slap_init( serverMode, serverName );
+	if ( rc ) {
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 18 );
 		goto destroy;
 	}
 
-	if ( slap_controls_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: controls initialization error\n", 0, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY,
-		    "controls initialization error\n",
-		    0, 0, 0 );
-#endif
-
-		goto destroy;
-	}
-
-#ifdef HAVE_TLS
-	/* Library defaults to full certificate checking. This is correct when
-	 * a client is verifying a server because all servers should have a
-	 * valid cert. But few clients have valid certs, so we want our default
-	 * to be no checking. The config file can override this as usual.
-	 */
-	rc = 0;
-	(void) ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_REQUIRE_CERT, &rc );
-#endif
-
-#ifdef LDAP_SLAPI
-	if ( slapi_int_initialize() != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, CRIT, "main: slapi initialization error\n", 0, 0, 0 );
-#else
-		Debug( LDAP_DEBUG_ANY,
-		    "slapi initialization error\n",
-		    0, 0, 0 );
-#endif
-
-		goto destroy;
-	}
-#endif /* LDAP_SLAPI */
-
-	if ( overlay_init() ) {
-		goto destroy;
-	}
-
-	if ( read_config( configfile, 0 ) != 0 ) {
+	if ( read_config( configfile, configdir ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 19 );
 
@@ -551,6 +675,21 @@ int main( int argc, char **argv )
 		goto destroy;
 	}
 
+	if ( debug_unknowns ) {
+		rc = parse_debug_unknowns( debug_unknowns, &slap_debug );
+		ldap_charray_free( debug_unknowns );
+		debug_unknowns = NULL;
+		if ( rc )
+			goto destroy;
+	}
+	if ( syslog_unknowns ) {
+		rc = parse_debug_unknowns( syslog_unknowns, &ldap_syslog );
+		ldap_charray_free( syslog_unknowns );
+		syslog_unknowns = NULL;
+		if ( rc )
+			goto destroy;
+	}
+
 	if ( check & CHECK_CONFIG ) {
 		fprintf( stderr, "config check succeeded\n" );
 
@@ -561,25 +700,18 @@ int main( int argc, char **argv )
 		}
 	}
 
-	if ( glue_sub_init( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: subordinate config error\n", 0, 0, 0 );
-#else
+	if ( glue_sub_attach( ) != 0 ) {
 		Debug( LDAP_DEBUG_ANY,
 		    "subordinate config error\n",
 		    0, 0, 0 );
-#endif
+
 		goto destroy;
 	}
 
 	if ( slap_schema_check( ) != 0 ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: schema prep error\n", 0, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "schema prep error\n",
 		    0, 0, 0 );
-#endif
 
 		goto destroy;
 	}
@@ -587,13 +719,9 @@ int main( int argc, char **argv )
 #ifdef HAVE_TLS
 	rc = ldap_pvt_tls_init();
 	if( rc != 0) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( SLAPD, CRIT, "main: tls init failed: %d\n", rc, 0, 0 );
-#else
 		Debug( LDAP_DEBUG_ANY,
 		    "main: TLS init failed: %d\n",
 		    0, 0, 0 );
-#endif
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
 		goto destroy;
@@ -608,23 +736,20 @@ int main( int argc, char **argv )
 		/* Force new ctx to be created */
 		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, NULL );
 
-		rc = ldap_pvt_tls_init_def_ctx();
-		if( rc != 0) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( SLAPD, CRIT, "main: tls init def ctx failed: %d\n", rc, 0, 0 );
-#else
+		rc = ldap_pvt_tls_init_def_ctx( 1 );
+		if( rc == 0 ) {
+			ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &slap_tls_ctx );
+			/* Restore previous ctx */
+			ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, def_ctx );
+			load_extop( &slap_EXOP_START_TLS, 0, starttls_extop );
+		} else if ( rc != LDAP_NOT_SUPPORTED ) {
 			Debug( LDAP_DEBUG_ANY,
 			    "main: TLS init def ctx failed: %d\n",
 			    rc, 0, 0 );
-#endif
 			rc = 1;
 			SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 20 );
 			goto destroy;
 		}
-		/* Retrieve slapd's own ctx */
-		ldap_pvt_tls_get_option( NULL, LDAP_OPT_X_TLS_CTX, &slap_tls_ctx );
-		/* Restore previous ctx */
-		ldap_pvt_tls_set_option( NULL, LDAP_OPT_X_TLS_CTX, def_ctx );
 	}
 #endif
 
@@ -658,56 +783,72 @@ int main( int argc, char **argv )
 	mal_leaktrace(1);
 #endif
 
+	if ( slapd_pid_file != NULL ) {
+		FILE *fp = fopen( slapd_pid_file, "w" );
+
+		if ( fp == NULL ) {
+			int save_errno = errno;
+
+			Debug( LDAP_DEBUG_ANY, "unable to open pid file "
+				"\"%s\": %d (%s)\n",
+				slapd_pid_file,
+				save_errno, strerror( save_errno ) );
+
+			free( slapd_pid_file );
+			slapd_pid_file = NULL;
+
+			rc = 1;
+			goto destroy;
+		}
+		fprintf( fp, "%d\n", (int) getpid() );
+		fclose( fp );
+		slapd_pid_file_unlink = 1;
+	}
+
+	if ( slapd_args_file != NULL ) {
+		FILE *fp = fopen( slapd_args_file, "w" );
+
+		if ( fp == NULL ) {
+			int save_errno = errno;
+
+			Debug( LDAP_DEBUG_ANY, "unable to open args file "
+				"\"%s\": %d (%s)\n",
+				slapd_args_file,
+				save_errno, strerror( save_errno ) );
+
+			free( slapd_args_file );
+			slapd_args_file = NULL;
+
+			rc = 1;
+			goto destroy;
+		}
+
+		for ( i = 0; i < g_argc; i++ ) {
+			fprintf( fp, "%s ", g_argv[i] );
+		}
+		fprintf( fp, "\n" );
+		fclose( fp );
+		slapd_args_file_unlink = 1;
+	}
+
 	/*
 	 * FIXME: moved here from slapd_daemon_task()
 	 * because back-monitor db_open() needs it
 	 */
 	time( &starttime );
 
-	if ( slap_startup( NULL )  != 0 ) {
+	if ( slap_startup( NULL ) != 0 ) {
 		rc = 1;
 		SERVICE_EXIT( ERROR_SERVICE_SPECIFIC_ERROR, 21 );
 		goto shutdown;
 	}
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, INFO, "main: slapd starting.\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "slapd starting\n", 0, 0, 0 );
-#endif
-
-
-	if ( slapd_pid_file != NULL ) {
-		FILE *fp = fopen( slapd_pid_file, "w" );
-
-		if( fp != NULL ) {
-			fprintf( fp, "%d\n", (int) getpid() );
-			fclose( fp );
-
-		} else {
-			free(slapd_pid_file);
-			slapd_pid_file = NULL;
-		}
-	}
-
-	if ( slapd_args_file != NULL ) {
-		FILE *fp = fopen( slapd_args_file, "w" );
-
-		if( fp != NULL ) {
-			for ( i = 0; i < g_argc; i++ ) {
-				fprintf( fp, "%s ", g_argv[i] );
-			}
-			fprintf( fp, "\n" );
-			fclose( fp );
-		} else {
-			free(slapd_args_file);
-			slapd_args_file = NULL;
-		}
-	}
 
 #ifdef HAVE_NT_EVENT_LOG
 	if (is_NT_Service)
-	lutil_LogStartedEvent( serverName, slap_debug, configfile, urls );
+	lutil_LogStartedEvent( serverName, slap_debug, configfile ?
+		configfile : SLAPD_DEFAULT_CONFIGFILE , urls );
 #endif
 
 	rc = slapd_daemon();
@@ -746,11 +887,7 @@ stop:
 	lutil_LogStoppedEvent( serverName );
 #endif
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( SLAPD, CRIT, "main: slapd stopped.\n", 0, 0, 0 );
-#else
 	Debug( LDAP_DEBUG_ANY, "slapd stopped.\n", 0, 0, 0 );
-#endif
 
 
 #ifdef HAVE_NT_SERVICE_MANAGER
@@ -772,15 +909,22 @@ stop:
 	ldap_pvt_tls_destroy();
 #endif
 
-	if ( slapd_pid_file != NULL ) {
+	if ( slapd_pid_file_unlink ) {
 		unlink( slapd_pid_file );
 	}
-	if ( slapd_args_file != NULL ) {
+	if ( slapd_args_file_unlink ) {
 		unlink( slapd_args_file );
 	}
 
 	config_destroy();
 
+	if ( configfile )
+		ch_free( configfile );
+	if ( configdir )
+		ch_free( configdir );
+	if ( urls )
+		ch_free( urls );
+
 #ifdef CSRIMALLOC
 	mal_dumpleaktrace( leakfile );
 #endif
